# coding=utf-8
# Copyright 2020-2021 The HuggingFace AutoTrain Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
# pylint: enable=line-too-long
import os


# ignore bnb warnings
os.environ["BITSANDBYTES_NOWELCOME"] = "1"
# os.environ["HF_HUB_DISABLE_PROGRESS_BARS"] = "1"
__version__ = "0.6.7"
